# -*- coding: utf-8 -*-
"""
Created on Tue Nov 20 14:36:08 2018

@author: sjoerd
"""

# -*- coding: utf-8 -*-#
#
# python script that uses the consistency excel file, cytometer data and quantiFlash to evaluate the PMTs,
# it will create SNR and DNR plots for all selected detectors

# date: 20 November 2018
# @author: Konrad von Volkmann, APE GmbH, Berlin
# @author: Stephan Benke, Cytometry Facility, University Zurich
# @author: Sebastian Falkowski, APE GmbH, Berlin
# @author: Sjoerd Visser, APE GmbH, Berlin

# -------------------------------------
# Don't change or delete this section
# -------------------------------------
# @inistart

# [script]
# name = 4. Eval. Beads

# [param1]
# name = Directory
# label = Data directory
# value = C:\Users\
# mandatory = true
# datatype = directory

# @iniend

'''
-------------------------------------------------------------------------------
Preperations
-------------------------------------------------------------------------------
'''
# Debugging
import pdb
import time

# Standard packages
import os
import numpy as np
import pandas as pd
import glob
import fcsparser
import configparser
import argparse
import re
import scipy.stats as sp

# Matplotlib packages
import matplotlib.pyplot as p
import matplotlib.patches as patches
from matplotlib.widgets import RectangleSelector
p.ioff()


# Excel packages?
import openpyxl
from openpyxl import Workbook
from openpyxl import load_workbook
from openpyxl import formatting, styles
from openpyxl.chart import  Reference, Series, ScatterChart
from openpyxl.chart.shapes import GraphicalProperties
from openpyxl.styles import  Border, Side, Font, Alignment
from openpyxl.utils import get_column_letter

import warnings
warnings.filterwarnings("ignore")

# ----------------------
# Functions
# ----------------------

def load_file(file):
    meta, df_data = fcsparser.parse(file, reformat_meta=True)
    return df_data

def init_excel(fname, sheets, titels):
    wb = Workbook()
    for nr, name in enumerate(sheets):
        wb.create_sheet(name)
        ws = wb[name]
        ws['A1'] = titels[nr]

        # Title banner from A1:Z1
        cell_range = ['%s1' % get_column_letter(nr) for nr in range(1,27)]
        # Banner style
        blue_fill = styles.PatternFill("solid", fgColor="4398ED")
        White_color_font = 'FFFFFF'
        White_font = styles.Font(size=28, bold=True, color=White_color_font)

        for i in cell_range:
            cell = ws[i]
            cell.fill = blue_fill
            cell.font = White_font

    wb.remove(wb['Sheet'])
    wb.save(fname)

'''
-------------------------------------------------------------------------------
Start of the main script
-------------------------------------------------------------------------------
'''

if __name__ == "__main__":
    global voltages, detectors, intensities

    print('------------------------------------')
    print('Script: 4: Evaluation of beads data ')
    print('------------------------------------')
    print('\n')


    # ----------------------
    # Load Data
    # ----------------------

    try:
        # Choosing data folder
        argparser = argparse.ArgumentParser(description='Description')
        argparser.add_argument('--Directory', metavar='Directory', type=str)
        args = argparser.parse_args()
        data_folder = args.Directory

        # Check for Evaluation folder, if not -> create
        directory = os.path.join(data_folder, 'Evaluation')
        if not os.path.exists(directory):
            os.makedirs(directory)
        if not os.path.exists(os.path.join(directory, 'Beads')):
            os.makedirs(os.path.join(directory, 'Beads'))

        print(data_folder)
        # Loading data in to xArrays
        print('\n --- Loading files --- \n')

    except Exception as e:
        print(e)
        print('==================')
        print('ERROR: Something went wrong loading data, check Data is correct')
        print('==================')

    # Check in folders for datafiles
    if os.path.isdir(os.path.join(data_folder, 'Beads')):
        Bead_files = glob.glob(os.path.join(data_folder, 'Beads', 'B*.fcs'))

    # del Bead_files[3]
    # del Bead_files[5]
    print('Bead files: %d' % np.size(Bead_files))

    # Determine the Voltages from file names
    try:
        voltages = [int(x[-7:-4]) for x in Bead_files]   # Voltages
    except Exception as e:
        print(e)
        print('==================')
        print('ERROR: Voltages could not be retreived from file names, are the filenames correctly formatted')
        print('==================')


    print('\n --- Evaluating  files --- \n')
    # Loop over all voltages
    for nr,V in enumerate(voltages):
        print('Evaluating V%d file' %V )
        print('#task=Evaluating V%d file' %V)
        print('#progress={:f}'.format(nr*100/np.size(voltages)))

        # Load ramp file
        df_beads = load_file(Bead_files[nr])

        # Drop certain channels
        detector_select='height'
        if detector_select == 'height':
            drop = 'A' # Only use H files
            drop_channels = [x for x in df_beads.columns if x[-1]==drop]
            df_beads.drop(drop_channels[2:], axis=1, inplace = True)
        if detector_select == 'area':
            drop = 'H' # Only use A files
            drop_channels = [x for x in df_beads.columns if x[-1]==drop]
            df_beads.drop(drop_channels, axis=1, inplace = True)
        if detector_select == 'both':
            drop_channels = []
            pass # use both files
        df_beads.drop('Time',axis=1, inplace=True)

        # def rect_select_callback(eclick, erelease):
        #     ''' Eclick and erelease are the press and release events'''
        #     global bounds
        #     bounds = [float(x) for x in RS.extents]



        # print('Select the singlets population for voltage: %d' %V)
        # fig, current_ax = plt.subplots()
        # Title_string='Select primary population for file: Voltage %d'%V
        # plt.plot(df_beads['FSC-A'],df_beads['SSC-A'],'o',markersize=0.7)
        # plt.title(Title_string)
        # plt.text(0.95,0.95,'Press x(close) when selection is made')
        # RS = RectangleSelector(current_ax, rect_select_callback,
        #                                 drawtype='box', useblit=True,
        #                                 button=[1, 3],  # don't use middle button
        #                                 minspanx=5, minspany=5,
        #                                 spancoords='pixels',
        #                                 interactive=True)
        # plt.show()

        # df_data_sel = df_beads.copy()
        # df_data_sel = df_beads[(df_beads['FSC-A']>bounds[0]) & (df_beads['FSC-A']<bounds[1]) & (df_beads['SSC-A']>bounds[2]) & (df_beads['SSC-A']<bounds[3])]
        # N = np.shape(df_data_sel)[0]
        # print('Selection contains %d' %N)
        # df_beads = df_data_sel.copy()
        # Evaluate Ramp file -> df_ramp_median_100npe and df_K_factor_mean
        gmean = pd.Series(sp.stats.gmean(df_beads,axis=0), index=df_beads.columns)
        gstd = pd.Series(df_beads.std())
        median = pd.Series(df_beads.median())

        if nr == 0:
        # First intensity, initialize dataframes
            df_gstd = pd.DataFrame(gstd)
            df_gmean = pd.DataFrame(gmean)
            df_median = pd.DataFrame(median)
        else:
            # Append/concat other data
            df_gstd = pd.concat([df_gstd, gstd], axis=1)
            df_gmean = pd.concat([df_gmean, gmean], axis=1)
            df_median = pd.concat([df_median, median], axis=1)

        channel_for_plot = df_beads.columns[4]
        histbins = int(df_beads.shape[0]/128)
        hist, bins = np.histogram(df_beads[channel_for_plot], bins=histbins)
        logbins = np.logspace(np.log10(bins[0]), np.log10(bins[-1]), len(bins))

        p.figure(1)
        p.clf()

        p.hist(df_beads[channel_for_plot],bins=logbins)
        p.xlim([0.1,2e5])
        p.xlabel(channel_for_plot)
        p.ylabel('counts [arb. units]')
        p.xscale('log')
        p.title('Beads measures at %d V' % V)
        data_dir, data_file = os.path.split(Bead_files[nr])
        data_file = data_file[:-3]+'png'
        plot_file = os.path.join(directory,'Beads',data_file)
        p.savefig(plot_file,facecolor='w',dpi=300)

    df_gstd.columns = voltages
    df_gmean.columns = voltages
    df_median.columns = voltages

    # Export to excel
    fname = os.path.join(directory, '2Peak.xlsx')

    # Initialization of excel
    sheets = ['gstd','gmean','median']
    titels = ['Geo std of peak 2 beads','Geo mean of peak 2 beads','Median of peak 2 beads']
    init_excel(fname, sheets, titels)

    wb = load_workbook(fname)
    writer = pd.ExcelWriter(fname, engine='openpyxl')
    writer.book = wb
    writer.sheets = dict((ws.title, ws) for ws in wb.worksheets)

    # Export Raw std
    offset=4
    df_gstd.to_excel(writer, sheet_name=sheets[0], startrow=offset)

    # Export Raw mean
    offset=4
    df_gmean.to_excel(writer, sheet_name=sheets[1], startrow=offset)

    # Export Raw median
    offset=4
    df_median.to_excel(writer, sheet_name=sheets[2], startrow=offset)

    wb.save(fname)
    wb.close()

    ''' add bead data to Results excel '''

    fname = os.path.join(directory,'Results.xlsx')

    wb = load_workbook(fname)
    writer = pd.ExcelWriter(fname, engine='openpyxl')
    writer.book = wb
    writer.sheets = dict((ws.title, ws) for ws in wb.worksheets)

    offset=18
    Detectors=df_gstd.index
    for nr,i in enumerate(writer.sheets):
        if i == 'Summary':
            pass
        else:

            # detector=Detectors[1+nr]
            ws = wb[i]
            detector=str(ws['A2'].value)
            # Read Excel file
            df_npe_R = df_gmean.loc[detector].copy()
            df_npe_LIB = df_gmean.loc[detector].copy()
            df_K_factor = df_gmean.loc[detector].copy()


            for nr2,i2 in enumerate(voltages):
                Letter =get_column_letter(nr2+2)

                cell_name = str(Letter) + '15'
                df_npe_R.loc[i2] = ws[cell_name].value

                cell_name = str(Letter) + '21'
                df_npe_LIB.loc[i2] = ws[cell_name].value

                cell_name = str(Letter) + '14'
                df_K_factor.loc[i2] = ws[cell_name].value

            df_npe_beads = df_gmean.loc[detector] * df_K_factor

            Black_font = styles.Font(size=16, bold=True)
            # Title
            ws['A28'] ='Beads'
            cell = ws['A28']
            cell.font = Black_font

            df_Beads = pd.concat([df_gmean.loc[detector], df_median.loc[detector], df_gstd.loc[detector], df_K_factor, df_npe_beads], axis=1)
            df_Beads.columns=['geo mean', 'Median', 'geo std', 'k_factor', 'Npe']
            df_Beads.T.to_excel(writer, sheet_name=i, startrow=28)

            df_SNR = 20*np.log10(df_npe_R/df_npe_LIB)
            df_DNR = 20*np.log10(df_K_factor/df_npe_LIB*262144)
            df_Beads = 20*np.log10(df_npe_beads/df_npe_LIB)

            ws['A36']='Results + Beads'
            cell = ws['A36']
            cell.font = Black_font

            df_beads_results = pd.concat([df_SNR, df_DNR, df_Beads], axis=1)
            df_beads_results.columns=['SNR', 'DNR', 'Beads']
            df_beads_results.T.to_excel(writer, sheet_name=i, startrow=36)


            # Plot
            c1 = ScatterChart()
            c1.title = detector
            c1.y_axis.title = 'Signal [dB]'
            c1.x_axis.title = 'Voltages [V]'
            c1.x_axis.scaling.min = np.min(voltages)-50
            c1.x_axis.scaling.max = np.max(voltages)+50
            c1.y_axis.scaling.min = 0
            c1.y_axis.scaling.max = 90
            c1.height = 12
            c1.width = 20
            offset = 36
            VOLTAGE = Reference(ws, min_col=2, max_col=1+np.size(voltages), min_row=offset+1, max_row=offset+1)

            SNR = Reference(ws, min_col=2, max_col=1+np.size(voltages), min_row=offset+2, max_row=offset+2)
            DNR = Reference(ws, min_col=2, max_col=1+np.size(voltages), min_row=offset+3, max_row=offset+3)
            Beads = Reference(ws, min_col=2, max_col=1+np.size(voltages), min_row=offset+4, max_row=offset+4)
            s1 = Series(DNR, xvalues=VOLTAGE, title="DNR")
            s2 = Series(SNR, xvalues=VOLTAGE, title="SNR")
            s3 = Series(Beads, xvalues=VOLTAGE, title="Beads")

            color1='e66101'
            s1.graphicalProperties.line.width = 50000
            s1.graphicalProperties.line.solidFill = color1          # Line Fill
            s1.marker.graphicalProperties.solidFill = color1        # Marker filling
            s1.marker.graphicalProperties.line.solidFill = color1   # Marker outline
            s1.marker.symbol = "circle"
            s1.marker.size = 10

            color2='fdb863'
            s2.graphicalProperties.line.width = 50000
            s2.graphicalProperties.line.solidFill = color2          # Line Fill
            s2.marker.graphicalProperties.solidFill = color2        # Marker filling
            s2.marker.graphicalProperties.line.solidFill = color2   # Marker outline
            s2.graphicalProperties.line.dashStyle = "sysDot"        # Dotted line
            s2.marker.symbol = "circle"
            s2.marker.size = 10

            color2='9400D3'
            s3.graphicalProperties.line.width = 50000
            s3.graphicalProperties.line.solidFill = color2          # Line Fill
            s3.marker.graphicalProperties.solidFill = color2        # Marker filling
            s3.marker.graphicalProperties.line.solidFill = color2   # Marker outline
            s3.graphicalProperties.line.dashStyle = "sysDash"        # Dash line
            s3.marker.symbol = "circle"
            s3.marker.size = 10

            c1.append(s1)
            c1.append(s2)
            c1.append(s3)

            ws.add_chart(c1, "P28")

    wb.save(fname)
    wb.close()



    print('\n')
    print('------------------------------------')
    print('DONE')
    print('------------------------------------')
    print('\n')
